# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Union

from langchain_community.utilities import GoogleSerperAPIWrapper

from neuro_san.interfaces.coded_tool import CodedTool

# Default parameters for google serper
K = 10  # number of search results
GL = "us"  # country
HL = "en"  # langauge
TYPE = "search"  # search type


class GoogleSerper(CodedTool):
    """
    CodedTool implementation which provides a way to do website search by Google Serper
    """

    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Union[Dict[str, Any], str]:
        """
        :param args: An argument dictionary whose keys are the parameters
                to the coded tool and whose values are the values passed for them
                by the calling agent.  This dictionary is to be treated as read-only.

                The argument dictionary expects the following keys:
                    "query" the query to search for.

        :param sly_data: A dictionary whose keys are defined by the agent hierarchy,
                but whose values are meant to be kept out of the chat stream.

                This dictionary is largely to be treated as read-only.
                It is possible to add key/value pairs to this dict that do not
                yet exist as a bulletin board, as long as the responsibility
                for which coded_tool publishes new entries is well understood
                by the agent chain implementation and the coded_tool implementation
                adding the data is not invoke()-ed more than once.

                Keys expected for this implementation are:
                    None

        :return:
            In case of successful execution:
                A dictionary including metadata.
            otherwise:
                a text string an error message in the format:
                "Error: <error message>"
        """
        # Get query from args
        query: str = args.get("query", "")
        if query == "":
            return "Error: No query provided."

        # Parameters for google serper

        # Country code to localize search results (e.g., "us" for United States)
        gl: str = args.get("gl", GL)
        # Language code for the search interface (e.g., "en" for English)
        hl: str = args.get("hl", HL)
        # Number of top search results to retrieve
        k: int = args.get("k", K)
        # Type of search (e.g., "news", "places", "images", or "search" for general)
        search_type: str = args.get("type", TYPE)
        # Search filter string (e.g., "qdr:d" for past day results); optional and can be used for time filtering
        # Default is None.
        tbs: str = args.get("tbs")

        # Create search with the above parameters
        search = GoogleSerperAPIWrapper(
            gl=gl,
            hl=hl,
            k=k,
            type=search_type,
            tbs=tbs
        )

        # Perform search asynchronously
        results = await search.aresults(query)

        return results




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from json.decoder import JSONDecodeError
from json_repair import loads

from neuro_san.internals.parsers.structure.structure_parser import StructureParser


class JsonStructureParser(StructureParser):
    """
    JSON implementation for a StructureParser.
    """

    def parse_structure(self, content: str) -> Dict[str, Any]:
        """
        Parse the single string content for any signs of structure

        :param content: The string to parse for structure
        :return: A dictionary structure that was embedded in the content.
                Will return None if no parseable structure is detected.
        """
        # Reset remainder on each call
        self.remainder = None

        meat: str = content
        delimiters: Dict[str, str] = {
            # Start : End
            "```json": "```",
            "```": "```",
            "`{": "}`",
            "{": "}",
        }

        for start_delim, end_delim in delimiters.items():
            if start_delim in content:

                # Note: This code assumes we only have one delimited JSON structure to parse
                #       within the content.

                # Well-formed per delimiter
                split_header: List[str] = content.split(start_delim)

                # Start the remainder off with everything before the json backtick business
                self.remainder = split_header[0]

                # Find the end of the backticks if any
                if end_delim != start_delim:
                    split_footer: List[str] = split_header[-1].split(end_delim)
                    meat = split_footer[0]
                    if len(split_footer) > 1:
                        # Add to the remainder anything outside the delimiting backticks
                        self.remainder += split_footer[-1]
                else:
                    meat = split_header[1]
                    if len(split_header) > 2:
                        # Add the remaining with the end delimiter.
                        # We are only parsing the first we find.
                        self.remainder += end_delim.join(split_header[2:])

                # Meat is everything in between, maybe with start and end delims on either end.
                meat = meat.strip()

                # Maybe add the delimiters back to help parsing the meat.
                use_delims: bool = start_delim != end_delim
                if use_delims:
                    meat = f"{start_delim}{meat}{end_delim}"

                break

        # Attempt parsing the structure from the meat
        structure: Dict[str, Any] = None
        try:
            structure = loads(meat)
            if not isinstance(structure, Dict):
                # json_repair seems to sometimes return an empty string if there is nothing
                # for it to grab onto.
                structure = None
        except JSONDecodeError:
            # Couldn't parse
            self.remainder = None

        # Strip any whitespace of the ends of any remainder.
        if self.remainder is not None:
            self.remainder = self.remainder.strip()

        return structure




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.journals.journal import Journal


class CompoundJournal(Journal):
    """
    A Journal implementation that can service multiple other Journal instances
    """

    def __init__(self, journals: List[Journal] = None):
        """
        Constructor

        :param journals: A List of Journal instances to simultaneously service
        """
        self.journals: List[Journal] = journals
        if self.journals is None:
            self.journals = []

    def add_journal(self, journal: Journal):
        """
        Adds a journal to the list
        :param journal: A Journal instance to service
        """
        self.journals.append(journal)

    async def write_message(self, message: BaseMessage, origin: List[Dict[str, Any]]):
        """
        Writes a BaseMessage entry into the journal
        :param message: The BaseMessage instance to write to the journal
        :param origin: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        for journal in self.journals:
            await journal.write_message(message, origin)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import List
from typing import Union

import json

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.messages.chat_message_type import ChatMessageType


class IntraAgentMessageUtils:
    """
    Utility class to parse messages with role information.
    This ancient style of message is used between agents and tool calls
    and is agnostic to OpenAI vs langchain.
    """

    @staticmethod
    def generate_response(the_messages: List[Any]) -> str:
        """
        :param the_messages: A list of messages. Could be OpenAI or langchain messages.
        :return: a JSON-ification of the list of messages.
        """
        response_list = []
        for index, one_message in enumerate(the_messages):
            # Duplicate the role message before every tool response role message
            role: str = IntraAgentMessageUtils._get_role(one_message)
            if role == "tool" and index > 0:
                previous_message = the_messages[index - 1]
                previous_role: str = IntraAgentMessageUtils._get_role(previous_message)
                if previous_role == "assistant":
                    new_message = {
                        "role": previous_role,
                        "content": IntraAgentMessageUtils.get_content(previous_message)
                    }
                    response_list.append(new_message)

            message_dict = {
                "role": role,
                "content": IntraAgentMessageUtils.get_content(one_message)
            }
            response_list.append(message_dict)

        return json.dumps(response_list)

    @staticmethod
    def _get_role(message: Any) -> str:
        """
        :param message: Either an OpenAI message or a langchain BaseMessage
        :return: A string describing the role of the message
        """

        if hasattr(message, "role"):
            return message.role

        # Check the look-up table above
        role: str = ChatMessageType.message_to_role(message)
        if role is not None:
            return role

        raise ValueError(f"Don't know how to handle message type {message.__class__.__name__}")

    @staticmethod
    def get_content(message: Any) -> str:
        """
        :param message: Either an OpenAI message or a langchain BaseMessage
        :return: A string describing the content of the message
        """

        if isinstance(message, BaseMessage):
            # For OpenAI and Ollama, content of AI message is a string but content from
            # Anthropic AI message can either be a single string or a list of content blocks.
            # If it is a list, "text" is a key of a dictionary which is the first element of
            # the list. For more details: https://python.langchain.com/docs/integrations/chat/anthropic/#content-blocks
            content: Union[str, List] = message.content
            if isinstance(content, list):
                content = content[0].get("text", "")
            return content

        if hasattr(message, "content"):
            if not any(message.content):
                return ""
            return message.content[0].text.value

        raise ValueError(f"Don't know how to handle message type {message.__class__.__name__}")




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""

from typing import Any
from typing import Dict
from typing import List

import random
import threading

from tornado.ioloop import IOLoop

from neuro_san.interfaces.concierge_session import ConciergeSession
from neuro_san.internals.network_providers.service_agent_network_storage import ServiceAgentNetworkStorage
from neuro_san.internals.network_providers.single_agent_network_provider import SingleAgentNetworkProvider
from neuro_san.service.generic.agent_server_logging import AgentServerLogging
from neuro_san.service.generic.async_agent_service import AsyncAgentService
from neuro_san.service.http.handlers.health_check_handler import HealthCheckHandler
from neuro_san.service.http.handlers.connectivity_handler import ConnectivityHandler
from neuro_san.service.http.handlers.function_handler import FunctionHandler
from neuro_san.service.http.handlers.streaming_chat_handler import StreamingChatHandler
from neuro_san.service.http.handlers.concierge_handler import ConciergeHandler
from neuro_san.service.http.handlers.openapi_publish_handler import OpenApiPublishHandler
from neuro_san.service.http.interfaces.agent_authorizer import AgentAuthorizer
from neuro_san.service.http.interfaces.agents_updater import AgentsUpdater
from neuro_san.service.http.logging.http_logger import HttpLogger
from neuro_san.service.http.server.http_server_app import HttpServerApp
from neuro_san.service.interfaces.agent_server import AgentServer
from neuro_san.service.interfaces.event_loop_logger import EventLoopLogger
from neuro_san.session.direct_concierge_session import DirectConciergeSession


class HttpSidecar(AgentAuthorizer, AgentsUpdater):
    """
    Class provides simple http endpoint for neuro-san API,
    working as a client to neuro-san gRPC service.
    """
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments, too-many-positional-arguments

    TIMEOUT_TO_START_SECONDS: int = 10

    def __init__(self, start_event: threading.Event,
                 port: int, http_port: int,
                 openapi_service_spec_path: str,
                 requests_limit: int,
                 forwarded_request_metadata: str = AgentServer.DEFAULT_FORWARDED_REQUEST_METADATA):
        """
        Constructor:
        :param start_event: event to await before starting actual service;
        :param port: port for gRPC neuro-san service;
        :param http_port: port for http neuro-san service;
        :param openapi_service_spec_path: path to a file with OpenAPI service specification;
        :param request_limit: The number of requests to service before shutting down.
                        This is useful to be sure production environments can handle
                        a service occasionally going down.
        :param forwarded_request_metadata: A space-delimited list of http metadata request keys
               to forward to logs/other requests
        """
        self.server_name_for_logs: str = "Http Server"
        self.start_event: threading.Event = start_event
        self.port = port
        self.http_port = http_port

        # Randomize requests limit for this server instance.
        # Lower and upper bounds for number of requests before shutting down
        if requests_limit == -1:
            # Unlimited requests
            self.requests_limit = -1
        else:
            request_limit_lower = round(requests_limit * 0.90)
            request_limit_upper = round(requests_limit * 1.10)
            self.requests_limit = random.randint(request_limit_lower, request_limit_upper)

        self.logger = None
        self.openapi_service_spec_path: str = openapi_service_spec_path
        self.forwarded_request_metadata: List[str] = forwarded_request_metadata.split(" ")
        self.allowed_agents: Dict[str, AsyncAgentService] = {}
        self.lock = None

    def __call__(self, other_server: AgentServer):
        """
        Method to be called by a thread running tornado HTTP server
        to actually start serving requests.
        """
        self.lock = threading.Lock()
        self.logger = HttpLogger(self.forwarded_request_metadata)
        app = self.make_app(self.requests_limit, self.logger)

        # Wait for "go" signal which will be set by gRPC server and corresponding machinery
        # when everything is ready for servicing.
        while not self.start_event.wait(timeout=self.TIMEOUT_TO_START_SECONDS):
            self.logger.error({},
                              "Timeout (%d sec) waiting for signal to HTTP server to start",
                              self.TIMEOUT_TO_START_SECONDS)

        app.listen(self.http_port)
        self.logger.info({}, "HTTP server is running on port %d", self.http_port)
        self.logger.info({}, "HTTP server is shutting down after %d requests", self.requests_limit)
        # Construct initial "allowed" list of agents:
        # no metadata to use here yet.
        self.update_agents(metadata={})
        self.logger.debug({}, "Serving agents: %s", repr(self.allowed_agents.keys()))

        IOLoop.current().start()
        self.logger.info({}, "Http server stopped.")
        if other_server is not None:
            other_server.stop()

    def make_app(self, requests_limit: int, logger: EventLoopLogger):
        """
        Construct tornado HTTP "application" to run.
        """
        request_data: Dict[str, Any] = self.build_request_data()
        health_request_data: Dict[str, Any] = {
            "forwarded_request_metadata": self.forwarded_request_metadata
        }
        handlers = []
        handlers.append(("/", HealthCheckHandler, health_request_data))
        handlers.append(("/healthz", HealthCheckHandler, health_request_data))
        handlers.append(("/api/v1/list", ConciergeHandler, request_data))
        handlers.append(("/api/v1/docs", OpenApiPublishHandler, request_data))

        # Register templated request paths for agent API methods:
        # regexp format used here is that of Python Re standard library.
        handlers.append((r"/api/v1/([^/]+)/function", FunctionHandler, request_data))
        handlers.append((r"/api/v1/([^/]+)/connectivity", ConnectivityHandler, request_data))
        handlers.append((r"/api/v1/([^/]+)/streaming_chat", StreamingChatHandler, request_data))

        return HttpServerApp(handlers, requests_limit, logger)

    def allow(self, agent_name) -> AsyncAgentService:
        return self.allowed_agents.get(agent_name, None)

    def update_agents(self, metadata: Dict[str, Any]):
        """
        Update list of agents for which serving is allowed.
        :param metadata: metadata to be used for logging if necessary.
        :return: nothing
        """
        data: Dict[str, Any] = {}
        session: ConciergeSession = DirectConciergeSession(metadata=metadata)
        agents_dict: Dict[str, List[Dict[str, str]]] = session.list(data)
        agents_list: List[Dict[str, str]] = agents_dict["agents"]
        agents: List[str] = []
        for agent_dict in agents_list:
            agents.append(agent_dict["agent_name"])
        with self.lock:
            # We assume all agents from "agents" list are enabled:
            for agent_name in agents:
                if self.allowed_agents.get(agent_name, None) is None:
                    self.add_agent(agent_name)
            # All other agents are disabled:
            allowed_set = set(self.allowed_agents.keys())
            for agent_name in allowed_set:
                if agent_name not in agents:
                    self.remove_agent(agent_name)

    def add_agent(self, agent_name: str):
        """
        Add agent to the map of known agents
        :param agent_name: name of an agent
        """
        agent_network_provider: SingleAgentNetworkProvider = \
            ServiceAgentNetworkStorage.get_instance().get_agent_network_provider(agent_name)
        # Convert back to a single string as required by constructor
        request_metadata_str: str = " ".join(self.forwarded_request_metadata)
        agent_server_logging: AgentServerLogging = \
            AgentServerLogging(self.server_name_for_logs, request_metadata_str)
        agent_service: AsyncAgentService = \
            AsyncAgentService(self.logger, None, agent_name, agent_network_provider, agent_server_logging)
        self.allowed_agents[agent_name] = agent_service

    def remove_agent(self, agent_name: str):
        """
        Remove agent from the map of known agents
        :param agent_name: name of an agent
        """
        self.allowed_agents.pop(agent_name, None)

    def build_request_data(self) -> Dict[str, Any]:
        """
        Build request data for Http handlers.
        :return: a dictionary with request data to be passed to a http handler.
        """
        return {
            "agent_policy": self,
            "agents_updater": self,
            "port": self.port,
            "forwarded_request_metadata": self.forwarded_request_metadata,
            "openapi_service_spec_path": self.openapi_service_spec_path
        }



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

import logging
import threading
from typing import Dict
from typing import List

from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.interfaces.agent_state_listener import AgentStateListener
from neuro_san.internals.network_providers.single_agent_network_provider import SingleAgentNetworkProvider


class ServiceAgentNetworkStorage:
    """
    Service-wide storage for AgentNetworkProviders.
    This class is a global singleton containing
    a table of currently active AgentNetworks for each agent registered to the service.
    Note: a mapping from an agent to its AgentNetwork is dynamic,
          as we can change agents definitions at service run-time.
    """

    instance = None

    def __init__(self):
        self.agents_table: Dict[str, AgentNetwork] = {}
        self.logger = logging.getLogger(self.__class__.__name__)
        self.lock = threading.Lock()
        self.listeners: List[AgentStateListener] = []

    @classmethod
    def get_instance(cls):
        """
        Get a singleton instance of this class
        """
        if not ServiceAgentNetworkStorage.instance:
            ServiceAgentNetworkStorage.instance = ServiceAgentNetworkStorage()
        return ServiceAgentNetworkStorage.instance

    def add_listener(self, listener: AgentStateListener):
        """
        Add a state listener to be notified when status of service agents changes.
        """
        self.listeners.append(listener)

    def remove_listener(self, listener: AgentStateListener):
        """
        Remove a state listener from registered set.
        """
        if listener in self.listeners:
            self.listeners.remove(listener)

    def add_agent_network(self, agent_name: str, agent_network: AgentNetwork):
        """
        This method is a single point of entry in the service
        where we register a new agent name -> AgentNetwork pair in the service scope
        or notify the service that for existing agent its AgentNetwork has been modified.
        """
        is_new: bool = False
        with self.lock:
            is_new = self.agents_table.get(agent_name, None) is None
            self.agents_table[agent_name] = agent_network
        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            if is_new:
                listener.agent_added(agent_name)
                self.logger.info("ADDED network for agent %s", agent_name)
            else:
                listener.agent_modified(agent_name)
                self.logger.info("REPLACED network for agent %s", agent_name)

    def setup_agent_networks(self, agent_networks: Dict[str, AgentNetwork]):
        """
        Replace agents networks with a new collection.
        Previous state could be empty.
        """
        current_agents = set(self.agents_table.keys())
        new_agents = set(agent_networks.keys())
        # Remove agents which are not in the new collection:
        agents_to_remove = current_agents - new_agents
        for agent_name in agents_to_remove:
            self.remove_agent_network(agent_name)
        # Now add (or possibly replace) agents from new collection:
        for agent_name in new_agents:
            self.add_agent_network(agent_name, agent_networks[agent_name])

    def remove_agent_network(self, agent_name: str):
        """
        Remove agent name and its AgentNetwork from service scope,
        so that agent becomes unavailable on our server.
        """
        with self.lock:
            self.agents_table.pop(agent_name, None)
        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            listener.agent_removed(agent_name)
        self.logger.info("REMOVED network for agent %s", agent_name)

    def get_agent_network_provider(self, agent_name: str) -> AgentNetworkProvider:
        """
        Get AgentNetworkProvider for a specific agent
        :param agent_name: name of an agent
        """
        return SingleAgentNetworkProvider(agent_name, self.agents_table)

    def get_agent_names(self) -> List[str]:
        """
        Return static list of agent names.
        """
        with self.lock:
            # Create static snapshot of agents names collection
            return list(self.agents_table.keys())




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from copy import copy

from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.internals.run_context.interfaces.agent_network_inspector import AgentNetworkInspector


class AgentNetwork(AgentNetworkInspector):
    """
    AgentNetworkInspector implementation for handling queries about a single
    agent network spec.  The data from the hocon file essentially lives here.
    """

    def __init__(self, config: Dict[str, Any], name: str):
        """
        Constructor

        :param config: The dictionary describing the entire agent network
        :param name: The name of the registry
        """
        self.config = config
        self.name = name
        self.agent_spec_map: Dict[str, Dict[str, Any]] = {}

        self.first_agent: str = None

        agent_specs = self.config.get("tools")
        if agent_specs is not None:
            for agent_spec in agent_specs:
                self.register(agent_spec)

    def get_config(self) -> Dict[str, Any]:
        """
        :return: The config dictionary passed into the constructor
        """
        return self.config

    def register(self, agent_spec: Dict[str, Any]):
        """
        :param agent_spec: A single agent to register
        """
        if agent_spec is None:
            return

        name: str = self.get_name_from_spec(agent_spec)
        if self.first_agent is None:
            self.first_agent = name

        if name in self.agent_spec_map:
            message: str = f"""
The agent named "{name}" appears to have a duplicate entry in its hocon file.
Agent names must be unique within the scope of a single hocon file.

Some things to try:
1. Rename one of the agents named "{name}". Don't forget to scrutinize all the
   tools references from other agents connecting to it.
2. If one definition is an alternate implementation, consider commenting out
   one of them with "#"-style comments.  (Yes, you can do that in a hocon file).
"""
            raise ValueError(message)

        self.agent_spec_map[name] = agent_spec

    def get_name_from_spec(self, agent_spec: Dict[str, Any]) -> str:
        """
        :param agent_spec: A single agent to register
        :return: The agent name as per the spec
        """
        extractor = DictionaryExtractor(agent_spec)
        name = extractor.get("function.name")
        if name is None:
            name = agent_spec.get("name")

        return name

    def get_agent_tool_spec(self, name: str) -> Dict[str, Any]:
        """
        :param name: The name of the agent tool to get out of the registry
        :return: The dictionary representing the spec registered agent
        """
        if name is None:
            return None

        return self.agent_spec_map.get(name)

    def find_front_man(self) -> str:
        """
        :return: A single tool name to use as the root of the chat agent.
                 This guy will be user facing.  If there are none or > 1,
                 an exception will be raised.
        """
        front_men: List[str] = []

        # Identify the "front-man" agent.
        # Primary heuristic: an agent with defined instructions and a function that takes no parameters.
        # The presence of instructions ensures it was explicitly defined, since users may add parameters
        # to front-men, making function signature alone unreliable.
        for name, agent_spec in self.agent_spec_map.items():
            instructions: str = agent_spec.get("instructions")
            function: Dict[str, Any] = agent_spec.get("function")
            if instructions is not None and function is not None and function.get("parameters") is None:
                front_men.append(name)

        if len(front_men) == 0:
            # The best way to find a front man is to see which agent was registered first
            front_men.append(self.first_agent)

        # Check for validity of our front-man candidates.
        valid_front_men: List[str] = copy(front_men)
        for front_man in front_men:

            # Check the agent spec of the front man for validity
            agent_spec: Dict[str, Any] = self.get_agent_tool_spec(front_man)

            if agent_spec.get("class") is not None:
                # Currently, front man cannot be a coded tool
                valid_front_men.remove(front_man)
            elif agent_spec.get("toolbox") is not None:
                # Currently, front man cannot from a toolbox
                valid_front_men.remove(front_man)

        if len(valid_front_men) == 0:
            raise ValueError("""
No front man found.
Here are some pre-conditions for an agent in your network to be a potential front man:
1) The agent's "function" does not have any "parameters" defined OR
2) The agent is the first listed among the "tools" of your agent hocon file

Disqualifiers. A front man cannot:
* be a CodedTool with a "class" definition
* be a tool with a "toolbox" definition
""")

        if len(valid_front_men) > 1:
            raise ValueError(f"Found > 1 front man for chat. Possibilities: {valid_front_men}")

        front_man: str = valid_front_men[0]

        return front_man

    def get_network_name(self) -> str:
        """
        :return: The network name of this AgentNetwork
        """
        return self.name
